@@ -1,5 +1,5 @@
 /*
- * m_ipt.c	iptables based targets 
+ * m_ipt.c	iptables based targets
  * 		utilities mostly ripped from iptables <duh, its the linux way>
  *
  *		This program is free software; you can distribute it and/or
@@ -7,11 +7,8 @@
  *		as published by the Free Software Foundation; either version
  *		2 of the License, or (at your option) any later version.
  *
- * Authors:  J Hadi Salim (hadi@cyberus.ca) 
- * 
- * TODO: bad bad hardcoding IPT_LIB_DIR and PROC_SYS_MODPROBE
- *
-*/
+ * Authors:  J Hadi Salim (hadi@cyberus.ca)
+ */
 
 #include <syslog.h>
 #include <sys/socket.h>
@@ -36,24 +33,9 @@
 #include <fcntl.h>
 #include <sys/wait.h>
 
-const char *pname = "tc-ipt";
-const char *tname = "mangle";
-const char *pversion = "0.1";
-
-#ifndef TRUE
-#define TRUE 1
-#endif
-#ifndef FALSE
-#define FALSE 0
-#endif
-
-#ifndef IPT_LIB_DIR
-#define IPT_LIB_DIR "/usr/local/lib/iptables"
-#endif
-
-#ifndef PROC_SYS_MODPROBE
-#define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
-#endif
+static const char *pname = "tc-ipt";
+static const char *tname = "mangle";
+static const char *pversion = "0.1";
 
 static const char *ipthooks[] = {
 	"NF_IP_PRE_ROUTING",
@@ -69,9 +51,11 @@ static struct option original_opts[] = {
 };
 
 static struct iptables_target *t_list = NULL;
+static struct option *opts = original_opts;
 static unsigned int global_option_offset = 0;
 #define OPTION_OFFSET 256
 
+char *lib_dir;
 
 void
 register_target(struct iptables_target *me)
@@ -125,7 +109,7 @@ addr_to_dotted(const struct in_addr *addrp)
 	return buf;
 }
 
-int string_to_number_ll(const char *s, unsigned long long min, 
+int string_to_number_ll(const char *s, unsigned long long min,
 			unsigned long long max,
 		 unsigned long long *ret)
 {
@@ -169,18 +153,13 @@ int string_to_number(const char *s, unsigned int min, unsigned int max,
 	return result;
 }
 
-static struct option *
-copy_options(struct option *oldopts)
+static void free_opts(struct option *opts)
 {
-	struct option *merge;
-	unsigned int num_old;
-	for (num_old = 0; oldopts[num_old].name; num_old++) ;
-	merge = malloc(sizeof (struct option) * (num_old + 1));
-	if (NULL == merge)
-		return NULL;
-	memcpy(merge, oldopts, num_old * sizeof (struct option));
-	memset(merge + num_old, 0, sizeof (struct option));
-	return merge;
+	if (opts != original_opts) {
+		free(opts);
+		opts = original_opts;
+		global_option_offset = 0;
+	}
 }
 
 static struct option *
@@ -231,14 +210,13 @@ find_t(char *name)
 }
 
 static struct iptables_target *
-get_target_name(char *name)
+get_target_name(const char *name)
 {
 	void *handle;
 	char *error;
 	char *new_name, *lname;
 	struct iptables_target *m;
-
-	char path[sizeof (IPT_LIB_DIR) + sizeof ("/libipt_.so") + strlen(name)];
+	char path[strlen(lib_dir) + sizeof ("/libipt_.so") + strlen(name)];
 
 	new_name = malloc(strlen(name) + 1);
 	lname = malloc(strlen(name) + 1);
@@ -269,10 +247,10 @@ get_target_name(char *name)
 		}
 	}
 
-	sprintf(path, IPT_LIB_DIR "/libipt_%s.so", new_name);
+	sprintf(path,  "%s/libipt_%s.so",lib_dir, new_name);
 	handle = dlopen(path, RTLD_LAZY);
 	if (!handle) {
-		sprintf(path, IPT_LIB_DIR "/libipt_%s.so", lname);
+		sprintf(path, lib_dir, "/libipt_%s.so", lname);
 		handle = dlopen(path, RTLD_LAZY);
 		if (!handle) {
 			fputs(dlerror(), stderr);
@@ -345,7 +323,7 @@ static void set_revision(char *name, u_int8_t revision)
 	name[IPT_FUNCTION_MAXNAMELEN - 1] = revision;
 }
 
-/* 
+/*
  * we may need to check for version mismatch
 */
 int
@@ -376,7 +354,7 @@ build_st(struct iptables_target *target, struct ipt_entry_target *t)
 	return -1;
 }
 
-static int parse_ipt(struct action_util *a,int *argc_p, 
+static int parse_ipt(struct action_util *a,int *argc_p,
 		     char ***argv_p, int tca_id, struct nlmsghdr *n)
 {
 	struct iptables_target *m = NULL;
@@ -385,7 +363,6 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 	int c;
 	int rargc = *argc_p;
 	char **argv = *argv_p;
-	struct option *opts;
 	int argc = 0, iargc = 0;
 	char k[16];
 	int res = -1;
@@ -394,6 +371,10 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 	__u32 hook = 0, index = 0;
 	res = 0;
 
+	lib_dir = getenv("IPTABLES_LIB_DIR");
+	if (!lib_dir)
+		lib_dir = IPT_LIB_DIR;
+
 	{
 		int i;
 		for (i = 0; i < rargc; i++) {
@@ -409,11 +390,6 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 		return -1;
 	}
 
-	opts = copy_options(original_opts);
-
-	if (NULL == opts)
-		return -1;
-
 	while (1) {
 		c = getopt_long(argc, argv, "j:", opts, NULL);
 		if (c == -1)
@@ -440,23 +416,14 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 		default:
 			memset(&fw, 0, sizeof (fw));
 			if (m) {
-				unsigned int fake_flags = 0;
 				m->parse(c - m->option_offset, argv, 0,
-					 &fake_flags, NULL, &m->t);
+					 &m->tflags, NULL, &m->t);
 			} else {
 				fprintf(stderr," failed to find target %s\n\n", optarg);
 				return -1;
 
 			}
 			ok++;
-
-			/*m->final_check(m->t); -- Is this necessary?
-			** useful when theres depencies
-			** eg ipt_TCPMSS.c has have the TCP match loaded
-			** before this can be used;
-			**  also seems the ECN target needs it 
-			*/
-
 			break;
 
 		}
@@ -466,6 +433,7 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 		if (matches(argv[optind], "index") == 0) {
 			if (get_u32(&index, argv[optind + 1], 10)) {
 				fprintf(stderr, "Illegal \"index\"\n");
+				free_opts(opts);
 				return -1;
 			}
 			iok++;
@@ -479,6 +447,10 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 		return -1;
 	}
 
+	/* check that we passed the correct parameters to the target */
+	if (m)
+		m->final_check(m->tflags);
+
 	{
 		struct tcmsg *t = NLMSG_DATA(n);
 		if (t->tcm_parent != TC_H_ROOT
@@ -517,8 +489,9 @@ static int parse_ipt(struct action_util *a,int *argc_p,
 	argv += optind;
 	*argc_p = rargc - iargc;
 	*argv_p = argv;
-	
+
 	optind = 1;
+	free_opts(opts);
 
 	return 0;
 
@@ -529,16 +502,10 @@ print_ipt(struct action_util *au,FILE * f, struct rtattr *arg)
 {
 	struct rtattr *tb[TCA_IPT_MAX + 1];
 	struct ipt_entry_target *t = NULL;
-	struct option *opts;
 
 	if (arg == NULL)
 		return -1;
 
-	opts = copy_options(original_opts);
-
-	if (NULL == opts)
-		return -1;
-
 	parse_rtattr_nested(tb, TCA_IPT_MAX, arg);
 
 	if (tb[TCA_IPT_TABLE] == NULL) {
@@ -597,10 +564,11 @@ print_ipt(struct action_util *au,FILE * f, struct rtattr *arg)
 				struct tcf_t *tm = RTA_DATA(tb[TCA_IPT_TM]);
 				print_tm(f,tm);
 			}
-		} 
+		}
 		fprintf(f, " \n");
 
 	}
+	free_opts(opts);
 
 	return 0;
 }
